unit SpiFilter;

{$IMPORTEDDATA OFF}

interface

uses Windows, spi_api, ImgFilter;

type
  TArchiveInfo = record
    FileName: WideString;
    FileSize: Cardinal;
    CompSize: Cardinal;
    DateTime: Int64;
  end;

  TArchive = class(TObject)
  private
    FPluginName: string;
    FPluginFileName: WideString;
    FFileName: string;
    FFileNameW: WideString;
    Handle: THandle;
    HInfo: HLOCAL;
    SpiFileInfoTop: PFileInfo;
    SpiFileInfo: PFileInfo;
    SpiFileInfoTopW: PFileInfoW;
    SpiFileInfoW: PFileInfoW;
    GetPluginInfo: TGetPluginInfo;
    GetArchiveInfo: TGetArchiveInfo;
    GetArchiveInfoW: TGetArchiveInfoW;
    GetFile: TGetFile;
    GetFileW: TGetFileW;
    function GetInfo(Position: Cardinal): TArchiveInfo;
  public
    constructor Create(const PluginFileName: WideString);
    destructor Destroy; override;
    procedure Open(const AFileName: WideString);
    function EachPosition(var Position: Cardinal): Boolean;
    procedure Extract(Position: Cardinal; Buf: PChar; BufSize: Cardinal);
    property FileName: string read FFileName;
    property Info[Position: Cardinal]: TArchiveInfo read GetInfo; default;
    property PluginName: string read FPluginName;
    property PluginFileName: WideString read FPluginFileName;
  end;

  TSpiFilter = class(TImgFilter)
  private
    FExtInfo: string;
    FFileName: WideString;
    GetPluginInfo: TGetPluginInfo;
    IsSupported: TIsSupported;
    GetPictureInfo: TGetPictureInfo;
    GetPicture: TGetPicture;
    ConfigurationDlg: TConfigurationDlg;
  public
    Handle: THandle;
    destructor Destroy; override;
    procedure LoadPlugin(const PluginFileName: WideString; UseExtInfo: Boolean);
    function Test(Buf: PChar; BufSize: Cardinal): Boolean; override;
    function TestSpi(const AFileName: string; Buf: PChar; BufSize: Cardinal):
        Boolean;
    procedure Decode(InBuf: PChar; InBufSize: Cardinal;
        var OutBuf: PChar; var OutBufSize: Cardinal; var ImgInfo: TImgInfo);
        override;
    function IsArchive: Boolean;
    function HasConfigDlg: Boolean;
    procedure ShowConfigDlg(hWnd: HWND);
    property FileName: WideString read FFileName;
    property ExtInfo: string read FExtInfo;
  end;


implementation

uses NLib, NSys, ifdib;

const
  SpiErrMsg: array [-1..10] of PChar = (
    'Not implemented',
    '',
    'User cancel',
    'Unknown format',
    'Broken data',
    'Cannot allocate memory',
    'Memory error',
    'File read error',
    'Cannot create window',
    'Internal error',
    'File write error',
    'End of file error'
  );

function ShortFileName(const S: WideString): string;
var
  len: DWORD;
  buf: WideString;
begin
  len := GetShortPathNameW(PWideChar(S), nil, 0);
  SetLength(buf, len-1);
  GetShortPathNameW(PWideChar(S), PWideChar(buf), len);
  Result := buf;
end;

procedure TSpiFilter.LoadPlugin(const PluginFileName: WideString;
    UseExtInfo: Boolean);
const
  maxexts = 100;
var
  buf: array [0..255] of Char;
  i: Integer;
begin
  Handle := LoadLibraryW(PWideChar(PluginFileName));
  if Handle <> 0 then
  begin
    FFileName := PluginFileName;
    GetPluginInfo := GetProcAddress(Handle, 'GetPluginInfo');
    IsSupported := GetProcAddress(Handle, 'IsSupported');
    GetPictureInfo := GetProcAddress(Handle, 'GetPictureInfo');
    GetPicture := GetProcAddress(Handle, 'GetPicture');
    ConfigurationDlg := GetProcAddress(Handle, 'ConfigurationDlg');
    if Assigned(GetPluginInfo) then
    begin
      GetPluginInfo(1, buf, SizeOf(buf));
      FName := buf;

      FExtInfo := '';
      if UseExtInfo then
        for i := 0 to maxexts-1 do
        begin
          if GetPluginInfo(2*i+2, buf, SizeOf(buf)) = 0 then
            Break;
          FExtInfo := FExtInfo + buf + ';';
        end;
    end;
  end;
end;

destructor TSpiFilter.Destroy;
begin
  if Handle <> 0 then
    FreeLibrary(Handle);
end;

function TSpiFilter.Test(Buf: PChar; BufSize: Cardinal): Boolean;
begin
  Result := TestSpi('', Buf, BufSize);
end;

function TSpiFilter.TestSpi(const AFileName: string; Buf: PChar;
    BufSize: Cardinal): Boolean;
var
  tmpbuf: array [0..2047] of Char;
  root, ext: WideString;
  exta, info: string;
begin
  if not Assigned(IsSupported) then
  begin
    Result := False;
    Exit;
  end;

  if ExtInfo <> '' then
  begin
    if AFileName = '' then // stdin
    begin
      Result := False;
      Exit;
    end;
    SplitExt(AFileName, root, ext);
    exta := StrUpper(ext);
    info := StrUpper(ExtInfo);
    if FindBytes(info[1], Length(info), '*'+exta+';') = -1 then
    begin
      Result := False;
      Exit;
    end;
  end;

  // Call IsSupported() with copyed buffer. IsSupprted() may destroy buffer.
  if BufSize > SizeOf(tmpbuf) then
    BufSize := SizeOf(tmpbuf);
  FillChar(tmpbuf, SizeOf(tmpbuf), 0);
  Move(Buf^, tmpbuf, BufSize);
  Result := IsSupported(PChar(AFileName), DWORD(@tmpbuf)) <> 0;
end;

function ProgressCallback(nNum, nDemon: Integer; lData : Longint): Integer; stdcall;
begin
  Result := 0;
end;

procedure SpiCheck(Code: Integer; const Name: string);
var
  msg: string;
begin
  if Code = 0 then Exit;

  case Code of
    Low(SpiErrMsg)..High(SpiErrMsg): msg := SpiErrMsg[Code];
  else
    msg := 'error ' + IStr(Code);
  end;
  raise EImgFilterError.Create(Name+': '+msg);
end;

procedure TSpiFilter.Decode(InBuf: PChar; InBufSize: Cardinal;
    var OutBuf: PChar; var OutBufSize: Cardinal; var ImgInfo: TImgInfo);
var
  hbinfo, hbm: HLOCAL;
  size: Integer;
  bmi: TMaxBitmapInfo;
  picinfo: TPictureInfo;
begin
  // For plugin which do not set all picinfo paramateres.
  FillChar(picinfo, SizeOf(picinfo), 0);
  try
    if Assigned(GetPictureInfo) then
      SpiCheck(GetPictureInfo(InBuf, InBufSize, 1, picinfo), 'GetPictureInfo');

    with ImgInfo do
    begin
      Bpp := picinfo.colorDepth;
      FormatName := '';
      if picinfo.hInfo = 0 then
        Comment := ''
      else begin
        SetString(Comment, PChar(LocalLock(picinfo.hInfo)),
          LocalSize(picinfo.hInfo)-1); // -1: for last null character
        LocalUnLock(picinfo.hInfo);
      end;
      Components := 3; // DIB decoded by SPI is always 24bpp(3bytes).
    end;
  finally
    if picinfo.hInfo <> 0 then
      LocalFree(picinfo.hInfo);
  end;

  if not Assigned(GetPicture) then
    raise EImgFilterError.Create('GetPicture not found');

  // Call GetPicture() with dummy ProgressCallback function.
  // Some plugin may cause error if ProgressCallback is nil.
  hbinfo := 0;
  hbm := 0;
  try
    SpiCheck(GetPicture(InBuf, InBufSize, 1, hbinfo, hbm, ProgressCallback, 0),
        'GetPicture');

    size := LocalSize(hbinfo);
    if size > SizeOf(TMaxBitmapInfo) then
      size := SizeOf(TMaxBitmapInfo);
    Move(PMaxBitmapInfo(LocalLock(hbinfo))^, bmi, size);
    LocalUnLock(hbinfo);

    ImgInfo.Width := bmi.bmiHeader.biWidth;
    ImgInfo.Height := Abs(bmi.bmiHeader.biHeight);

    OutBufSize := ((ImgInfo.Width * 3 + 3) and not 3) * ImgInfo.Height;
    OutBuf := Alloc(OutBufSize);
    if OutBuf = nil then
      raise EImgFilterError.Create('output buffer allocation failed');

    DecodeDIB(LocalLock(hbm), LocalSize(hbm), bmi, OutBuf, OutBufSize);
    LocalUnLock(hbm);
  finally
    if hbm <> 0 then
      LocalFree(hbm);
  end;
end;

function TSpiFilter.IsArchive: Boolean;
var
  ver: array [0..15] of Char;
begin
  FillChar(ver, SizeOf(ver), 0);
  if Assigned(GetPluginInfo) then
    GetPluginInfo(0, ver, SizeOf(ver));
  Result := Copy(ver, 1, 4) = '00AM';
end;

function TSpiFilter.HasConfigDlg: Boolean;
begin
  Result := Assigned(ConfigurationDlg);
end;

procedure TSpiFilter.ShowConfigDlg(hWnd: HWND);
begin
  SpiCheck(ConfigurationDlg(hWnd, 1), 'ConfigurationDlg');
end;


{ TArchive }

constructor TArchive.Create(const PluginFileName: WideString);
var
  buf: array [0..255] of Char;
begin
  Handle := LoadLibraryW(PWideChar(PluginFileName));
  if Handle <> 0 then
  begin
    FPluginFilename := PluginFileName;
    GetArchiveInfo := GetProcAddress(Handle, 'GetArchiveInfo');
    GetArchiveInfoW := GetProcAddress(Handle, 'GetArchiveInfoW');
    GetFile := GetProcAddress(Handle, 'GetFile');
    GetFileW := GetProcAddress(Handle, 'GetFileW');
    GetPluginInfo := GetProcAddress(Handle, 'GetPluginInfo');
    if Assigned(GetPluginInfo) then
    begin
      GetPluginInfo(1, buf, SizeOf(buf));
      FPluginName := buf;
    end;
  end;
end;

procedure TArchive.Open(const AFileName: WideString);
begin
  SpiFileInfoTop := nil;
  SpiFileInfo := nil;
  SpiFileInfoTopW := nil;
  SpiFileInfoW := nil;
  if Assigned(GetArchiveInfoW) then begin
    FFileNameW := AFileName;
    SpiCheck(GetArchiveInfoW(PWideChar(FFileNameW), 0, 0, HInfo), 'GetArchiveInfoW');
    SpiFileInfoTopW := LocalLock(HInfo);
    SpiFileInfoW := SpiFileInfoTopW;
  end else begin
    FFileName := ShortFileName(AFileName);
    if not Assigned(GetArchiveInfo) then
      raise EImgFilterError.Create('GetArchiveInfo not found');
    SpiCheck(GetArchiveInfo(PChar(FFileName), 0, 0, HInfo), 'GetArchiveInfo');
    SpiFileInfoTop := LocalLock(HInfo);
    SpiFileInfo := SpiFileInfoTop;
  end;
end;

destructor TArchive.Destroy;
begin
  if HInfo <> 0 then
  begin
    LocalUnlock(HInfo);
    LocalFree(HInfo);
  end;
  if Handle <> 0 then
    FreeLibrary(Handle);
end;

function TArchive.GetInfo(Position: Cardinal): TArchiveInfo;
var
  spiinfo: PFileInfo;
  spiinfow: PFileInfoW;
begin
  if SpiFileInfoTopW = nil then begin
    spiinfo := SpiFileInfoTop;
    while spiinfo.method[0] <> #0 do
      if spiinfo.position = Position then begin
        Result.FileName := spiinfo.path + spiinfo.filename;
        Result.FileSize := spiinfo.filesize;
        Result.CompSize := spiinfo.compsize;
        Result.DateTime := spiinfo.timestamp;
        Exit;
      end else
        Inc(spiinfo);
  end else begin
    spiinfow := SpiFileInfoTopW;
    while spiinfow.method[0] <> #0 do
      if spiinfow.position = Position then begin
        Result.FileName := WideString(spiinfow.path) + WideString(spiinfow.filename);
        Result.FileSize := spiinfow.filesize;
        Result.CompSize := spiinfow.compsize;
        Result.DateTime := spiinfow.timestamp;
        Exit;
      end else
        Inc(spiinfow);
  end;
  Result.FileName := '';
end;

function TArchive.EachPosition(var Position: Cardinal): Boolean;
begin
  if SpiFileInfoTopW = nil then begin
    Result := SpiFileInfo.method[0] <> #0;
    if Result then
      Position := SpiFileInfo.position;
    Inc(SpiFileInfo);
  end else begin
    Result := SpiFileInfoW.method[0] <> #0;
    if Result then
      Position := SpiFileInfoW.position;
    Inc(SpiFileInfoW);
  end;
end;

procedure TArchive.Extract(Position: Cardinal; Buf: PChar; BufSize: Cardinal);
var
  hdst: HLOCAL;
  p: PChar;
  fsize: UINT;
begin
  if Assigned(GetFileW) then
    SpiCheck(
        GetFileW(PWideChar(FFileNameW), Position, PWideChar(@hdst), $100, nil,0),
        'GetFileW')
  else begin
    if not Assigned(GetFile) then
      raise EImgFilterError.Create('GetFile not found');
    SpiCheck(
        GetFile(PChar(FFileName), Position, PChar(@hdst), $100, nil,0),
        'GetFile');
  end;
  p := LocalLock(hdst);
  fsize := LocalSize(hdst);
  if fsize > BufSize then
    fsize := BufSize;
  Move(p^, Buf^, fsize);
  LocalUnLock(hdst);
  LocalFree(hdst);
end;

end.
